!---------------------------------------------------------------------------------------
! PREP_FLUXES: read_ncdf
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> read_ncdf
!! 
!! Purpose:    Reads 3D data from a ncdf file.
!!
!! Interface:
!!
!!    Inputs
!!             filename  -  name of ncdf file
!!             varname   -  name of 3D floating variable
!!             lonname   -  name of longitude variable
!!             latname   -  name of latitude variable
!!             timename  -  name of time variable
!!             indextime -  index to start from in time dimension
!!             time      -  time dimension
!!             lon       -  longitude dimension (zero vector on input)
!!             lat       -  latitude dimension (zero vector on input)
!!             flx       -  fluxes (zero array on input)
!!
!---------------------------------------------------------------------------------------

subroutine read_ncdf(filename, varname, lonname, latname, timename,&
                     indextime, time, lon, lat, flx)

  use mod_var
  use netcdf

  implicit none

  character(len=max_path_len),        intent(in)     :: filename
  character(len=max_name_len),        intent(in)     :: varname, lonname, latname, timename
  integer,                            intent(in)     :: indextime
  real, dimension(nx_in),             intent(in out) :: lon
  real, dimension(ny_in),             intent(in out) :: lat
  real, dimension(ntime),             intent(in out) :: time
  real, dimension(nx_in,ny_in,ntime), intent(in out) :: flx

  integer :: nid, ierr, xid, yid, zid, ix, jy, kt
  integer :: lonid, latid, tid, varid
  integer :: nlon, nlat, nt
  real    :: missval

  print*, 'indextime = ',indextime

  ! open netcdf file
  write(*,*) 'Reading file: '//trim(filename)
  ierr=nf90_open(trim(filename),nf90_NOWRITE,nid)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: opening file ',trim(filename)
    stop
  endif

  ! inquire about variables
  ierr=nf90_inq_dimid(nid,trim(lonname),xid)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading dimension ',trim(lonname)
    stop
  endif
  ierr=nf90_inq_dimid(nid,trim(latname),yid)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading dimension ',trim(latname)
    stop
  endif
  if(trim(timename).ne.'none') then
    ierr=nf90_inq_dimid(nid,trim(timename),zid)
    if(ierr.ne.0) then
      write(*,*) 'ERROR: reading dimension: ',trim(timename)
      stop
    endif
  endif

  ! read dimension sizes
  ierr=nf90_inquire_dimension(nid,xid,len=nlon)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading dimension length'
    stop
  endif
  ierr=nf90_inquire_dimension(nid,yid,len=nlat)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading dimension length'
    stop
  endif
  if(trim(timename).ne.'none') then
    ierr=nf90_inquire_dimension(nid,zid,len=nt)
    if(ierr.ne.0) then
      write(*,*) 'ERROR: reading dimension length'
      stop
    endif
  else
    nt = 1
  endif

  ! check dimensions agree
!  if(nlon.ne.nx_in) then
!    write(*,*) 'ERROR: problem longitude dimension'
!    stop
!  endif
!  if(nlat.ne.ny_in) then
!    write(*,*) 'ERROR: problem latitude dimension'
!    stop
!  endif
!  if(nt.ne.ntime) then
!    write(*,*) 'ERROR: problem time dimension'
!    stop
!  endif

  ! read dimensions
  ierr=nf90_inq_varid(nid,lonname,lonid)
  ierr=nf90_get_var(nid,lonid,lon)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading variable longitude'
    stop
  endif
  ierr=nf90_inq_varid(nid,latname,latid)
  ierr=nf90_get_var(nid,latid,lat)
  if(ierr.ne.0) then
    write(*,*) 'ERROR: reading variable latitude'
    stop
  endif
  if(trim(timename).ne.'none') then
    ierr=nf90_inq_varid(nid,timename,tid)
    ierr=nf90_get_var(nid,tid,time(:),start=(/indextime/),count=(/ntime/))
    if(ierr.ne.0) then
      write(*,*) 'ERROR: reading variable time'
      stop
    endif
  else
    time = 1
  endif

  print*, 'time = ',time

  ! read variable
  ierr=nf90_inq_varid(nid,trim(varname),varid)
  ierr=nf90_get_var(nid,varid,flx(:,:,:),start=(/1,1,indextime/),count=(/nlon,nlat,ntime/))
  if(ierr.ne.0) then
    write(*,*) 'ERROR reading variable ',trim(varname)
    stop
  endif
 
  ! read fill value attribute
  ierr=nf90_get_att(nid,varid,'_FillValue',missval)
  if(ierr.eq.0) then
    ! variable has a _FillValue attribute
    where (flx.eq.missval) flx = 0.
  endif
  ierr=nf90_get_att(nid,varid,'missing_value',missval)
  if(ierr.eq.0) then
    ! variable has a missing_value attribute
    where (flx.eq.missval) flx = 0.
  endif

  ierr=nf90_close(nid)

end subroutine read_ncdf

