!---------------------------------------------------------------------------------------
! PREP_REGIONS: prep_regions
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> prep_regions 
!!
!! Purpose:    Calculates the grid aggregation based on the flux sensitivites from
!!             FLEXPART and writes the mapping operator to ncdf file. (See map_grid.f90
!!             for details on the aggregation. 
!!
!---------------------------------------------------------------------------------------

program prep_regions

  use mod_var
  use mod_ncdf
  use mod_settings

  implicit none

  ! set types
  type (config_t)                         :: config
  type (files_t)                          :: files
  
  character(max_path_len)                 :: settings_files
  character(max_path_len)                 :: settings_config
  character(max_path_len)                 :: settings_regions
  character(max_path_len)                 :: filename
  character(max_name_len)                 :: varname
  real, dimension(:,:), allocatable       :: surfinf
  real, dimension(:,:), allocatable       :: flux, mask
  real(kind=8), dimension(:), allocatable :: obstimes, avetimes
  real, dimension(:), allocatable         :: cdryair
  character(recname_len), dimension(:), allocatable :: recs
  real, dimension(:,:), allocatable       :: nbox_xy
  integer                                 :: ierr,ix,jy
  logical                                 :: lexist
  character(len=20)                       :: rowfmt
  

  call getarg(1,settings_files)
  if (settings_files == '') then
    stop 'ERROR: need to specify SETTINGS_files file'
  endif
 
  call getarg(2,settings_config)
  if (settings_files == '') then
    stop 'ERROR: need to specify SETTINGS_config file'
  endif

  call getarg(3,settings_regions)
  if (settings_regions == '') then
    stop 'ERROR: need to specify SETTINGS_regions file'
  endif

  write(*,*) 'Using: '//trim(settings_files)
  write(*,*) 'Using: '//trim(settings_config)
  write(*,*) 'Using: '//trim(settings_regions)

  ! read file settings
  call read_file_settings(settings_files, files)

  ! read config settings
  call read_config_settings(settings_config, config)

  ! read region settings
  call read_region_settings(settings_regions, files)

  ! initialize parameters
  call initialize(files, config)

  ! read land-sea mask 
  call read_lsm(files)

  ! read prior fluxes
  allocate( flux(nxregrid,nyregrid), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  filename = trim(files%path_prior)//trim(files%filename_flx)
  varname = files%varname_flx 
  call read_ncdf(files, config, filename, varname, flux)

  ! read observations
  allocate( obstimes(maxobs), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  allocate( avetimes(maxobs), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  allocate( recs(maxobs), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  allocate( cdryair(maxobs), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  obstimes(:) = 0d0
  avetimes(:) = 0d0
  call read_obs(files, config, recs, cdryair, obstimes, avetimes)

  ! mean surface influence (i.e. flux sensitivity)
  allocate( surfinf(nxregrid,nyregrid), stat=ierr )
  if ( ierr.ne.0 ) stop 'ERROR: not enough memory'
  surfinf(:,:) = 0.
  inquire ( file = trim(files%path_output)//'surfinf.nc', exist=lexist )
  if ( lexist ) then
    ! read surfinf from file
    filename = trim(files%path_output)//'surfinf.nc'
    varname = 'surfinf'
    call read_ncdf(files, config, filename, varname, surfinf)
  else
    call get_surfinf(files, config, recs, cdryair, obstimes, avetimes, surfinf)
  endif

  ! calculate regions
  allocate( nbox_xy(nxregrid,nyregrid), stat=ierr )
  call map_grid(config, flux, surfinf, nbox_xy)

  ! write regions file
  filename = files%file_regions
  varname = 'regions'
  call write_ncdf(files, filename, varname, nbox_xy)

  ! write the average fooptrint for all receptors
  open(50,file=trim(files%path_output)//'all_footprint_average.txt')
  do ix=1,nxregrid
      do jy=1,nyregrid
          write(50,*)  reg_lon(ix), reg_lat(jy), surfinf(ix,jy)
       end do
  end do
  close(50)

  ! write lsm 
  allocate( mask(nxregrid,nyregrid), stat=ierr)
  mask=lsm
  open(100,file=trim(files%path_output)//'lsm.txt',action='write',status='replace')
  write(rowfmt,'(A,I6,A)') '(',nxregrid,'(F5.2,1X))'
  do jy = 1,nyregrid
    write(100,rowfmt) lsm(:,jy)
  end do
  close(100)

end program



