!---------------------------------------------------------------------------------------
! FLEXINVERT: init_ghg
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> init_ghg
!! Initialize GHG variables
!!
!! Purpose:    Reads prior fluxes and observations of a GHG and calculates the 
!!             state and observation vectors as well as their uncertainties.
!!
!! Interface:
!!
!!    Inputs
!!             files   -  file data structure
!!             config  -  configuration settings data structure
!!             fluxes  -  flux data structure
!!             obs     -  observation data structure
!!             covar   -  error covariance data structure
!!
!!    Externals
!!             alloc_fluxes
!!             alloc_states
!!             caldate     
!!             read_ncdf        
!!             read_bg       
!!             correl
!!             error_cov        
!!             read_obs
!!             assign_obs
!!             init_cini
!!             perturb_state
!!             perturb_obs
!!
!---------------------------------------------------------------------------------------

subroutine init_ghg(files, config, fluxes, obs, states, covar)

  use mod_var
  use mod_settings
  use mod_dates
  use mod_tools
  use mod_strings
  use mod_fluxes
  use mod_obs
  use mod_states
  use mod_covar
  use mod_perturb
  use mod_save, only : save_for_restart 

  implicit none

  type (files_t),        intent (in)         :: files
  type (config_t),       intent (in)         :: config
  type (fluxes_t),       intent (in out)     :: fluxes
  type (obs_t),          intent (in out)     :: obs
  type (states_t),       intent (in out)     :: states
  type (covar_t),        intent (in out)     :: covar

  character(max_path_len)                    :: filename
  character(max_path_len)                    :: rowfmt
  character(len=4)                           :: adate
  character(len=2)                           :: amonth
  logical                                    :: lexist
  real, dimension(:,:,:), allocatable        :: flx, flxghg
  real, dimension(:,:), allocatable          :: datain, dataout
  real, dimension(nbox,ntstate)              :: flx_box
  real, dimension(nbox,ntstate)              :: err_box
  real, dimension(nbox)                      :: xerr
  real, dimension(nbox,nbox)                 :: corr
  integer                                    :: yyyymmdd, hhmiss
  real                                       :: area
  real(kind=8)                               :: jd
  integer                                    :: ix, jy, ix0, jy0, ix1, ix2, jy1, jy2
  integer                                    :: i, n, nb, nread, num, eomday
  integer                                    :: ierr
  real(kind=8), dimension(:), allocatable    :: jdate, timeghg
  character(len=recname_len), dimension(:), allocatable:: recs
  real                                       :: conc
  character(len=20)                          :: dump
  real, parameter                            :: smallnum=1.e-15

  ! global prior fluxes
  ! -------------------

  ! if nested run global fluxes are only used to calculate the contribution to  mixing ratios 
  ! from fluxes outside the inversion domain, otherwise, global fluxes are used to
  ! calculate the contribution to mixing ratios and as prior fluxes inside the domain.

  call alloc_fluxes(config, fluxes)
  fluxes%flx(:,:,:) = 0.
  fluxes%flxocn(:,:,:) = 0.

  ! loop over prior flux files

  allocate( flxghg(nxgrid,nygrid,max(1,(nday*24/nt_flx))), stat = ierr )
  if ( ierr.ne.0 ) stop 'ERROR init_ghg: not enough memory'
  allocate( timeghg(max(1,nday*24/nt_flx)) )

  n = 0
  jd = juldatei
  do while ( jd.le.juldatef )
    call caldate(jd, yyyymmdd, hhmiss)
    write(adate,'(I4.4)') yyyymmdd/10000
    write(amonth,'(I2.2)') (yyyymmdd-(yyyymmdd/10000)*10000)/100
    eomday = calceomday(yyyymmdd/100)    
    nread = eomday*24/nt_flx
    if ( nread.eq.0 ) nread = 1
    print*, 'init_ghg: nread = ',nread
    allocate( flx(nxgrid,nygrid,nread), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_ghg: not enough memory'
    filename = str_replace(files%filename_flx, 'YYYY', adate)
    if ( index(filename, 'MM').ne.0 ) then
      filename = str_replace(filename, 'MM', amonth)
    endif
    filename = trim(files%path_prior)//trim(filename)
    write(logid,*) 'Reading prior fluxes :'//trim(filename)
    inquire(file=trim(filename),exist=lexist)
    if (.not.lexist) then
      write(logid,*) 'ERROR: cannot find '//trim(filename)
      stop
    endif
    call read_ncdf(filename, &
                   files%varname_flx, &
                   files%lonname_flx, &
                   files%latname_flx, &
                   files%timename_flx,&
                   llx, lly, &
                   nxgrid, nygrid, &
                   jd, nread, num, &
                   flx)
    do i = 1, nread
      if ( ((n+i).gt.(nday*24/nt_flx)).and.(nday*24/nt_flx.ne.0) ) exit
      flxghg(:,:,n+i) = flx(:,:,i)
      timeghg(n+i) = jd + dble((i-1)*nt_flx)/24d0
    end do
    deallocate( flx )
    n = n + nread
    jd = jd + dble(eomday)
  end do
  n = min(n, nday*24/nt_flx)
  n = max(n, 1)
  print*, 'init_ghg: n = ',n
  print*, 'init_ghg: timeghg = ',timeghg

  ! convert to kg/m2/s
  flxghg = flxghg/3600.
  ! apply numerical scaling
  flxghg = flxghg*numscale

  ! average/interpolate to state vector timestamp
  allocate( datain(nxgrid,n) )
  allocate( dataout(nxgrid,ntstate) )
  do i = 1, ntstate
    fluxes%time(i) = juldatei + dble(i-1)*dble(statres)
  enddo
  if ( statres.gt.real(nt_flx/24) ) then
    ! average flux
    do jy = 1, nygrid
      datain = flxghg(:,jy,:)
      call average(nxgrid, n, timeghg, datain, ntstate, fluxes%time, dataout)
      fluxes%flx(:,jy,:) = dataout(:,:)
    end do
  else if ( statres.lt.real(nt_flx/24) ) then
    ! interpolate flux
    do jy = 1, nygrid
      datain = flxghg(:,jy,:)
      call interp(nxgrid, n, timeghg, datain, ntstate, fluxes%time, dataout)
      fluxes%flx(:,jy,:) = dataout(:,:)
    end do
  else if ( statres.eq.real(nt_flx/24) ) then
    fluxes%flx(:,:,:) = flxghg
  endif
  deallocate( datain )
  deallocate( dataout )
  deallocate( flxghg )
  deallocate( timeghg )

  ! ocean fluxes for lognormal case when inc_ocean is false
  ! cover only inversion domain
  if ( config%lognormal.and..not.config%inc_ocean.and..not.config%nested ) then
    ! use land-sea mask to set all land fluxes to zero
    ix1 = int((rllx - llx + dx)/dx)
    ix2 = int((rurx - llx)/dx)
    jy1 = int((rlly - lly + dy)/dy)
    jy2 = int((rury - lly)/dy)
    do n = 1, ntstate
      fluxes%flxocn(:,:,n) = abs(lsm(:,:) - 1.)*fluxes%flx(ix1:ix2,jy1:jy2,n)
    end do
  endif

  ! regional prior fluxes
  ! ---------------------

  if ( config%nested ) then

    allocate( flxghg(nxregrid,nyregrid,(max(1,nday*24/nt_flx))), stat = ierr )
    if ( ierr.ne.0 ) stop 'ERROR init_ghg: not enough memory'
    allocate( timeghg(max(1,nday*24/nt_flx)) )

    n = 0
    jd = juldatei
    do while ( jd.le.juldatef )
      call caldate(jd, yyyymmdd, hhmiss)
      write(adate,'(I4.4)') yyyymmdd/10000
      write(amonth,'(I2.2)') (yyyymmdd-(yyyymmdd/10000)*10000)/100
      eomday = calceomday(yyyymmdd/100)
      nread = eomday*24/nt_flx
      if ( nread.eq.0 ) nread = 1
      print*, 'init_ghg: reg nread = ',nread
      allocate( flx(nxregrid,nyregrid,nread), stat = ierr )
      if ( ierr.ne.0 ) stop 'ERROR init_ghg: not enough memory'
      filename = str_replace(files%filenest_flx, 'YYYY', adate)
      if ( index(filename, 'MM').ne.0 ) then
        filename = str_replace(filename, 'MM', amonth)
      endif
      filename = trim(files%path_prior)//trim(filename)
      write(logid,*) 'Reading prior fluxes :'//trim(filename)
      inquire(file=trim(filename),exist=lexist)
      if (.not.lexist) then
        write(logid,*) 'ERROR: cannot find '//trim(filename)
        stop
      endif
      call read_ncdf(filename, &
                     files%varnest_flx, &
                     files%lonnest_flx, &
                     files%latnest_flx, &
                     files%timenest_flx,&
                     rllx, rlly, &
                     nxregrid, nyregrid, &
                     jd, nread, num, &
                     flx)
      do i = 1, nread
        if ( ((n+i).gt.(nday*24)/nt_flx).and.((nday*24)/nt_flx.ne.0) ) exit
        timeghg(n+i) = jd + dble((i-1)*nt_flx)/24d0
        flxghg(:,:,n+i) = flx(:,:,i)
      end do
      deallocate( flx )
      n = n + nread
      jd = jd + dble(eomday)
    end do
    n = min(n, (nday*24)/nt_flx)
    n = max(n, 1)
    print*, 'init_ghg: reg n = ',n
    print*, 'init_ghg: reg timeghg = ',timeghg

    ! convert to kg/m2/s
    flxghg = flxghg/3600.
    ! apply numerical scaling
    flxghg = flxghg*numscale

    ! average/interpolate to state vector timestamp
    allocate( datain(nxregrid,n) )
    allocate( dataout(nxregrid,ntstate) )
    if ( statres.gt.real(nt_flx/24) ) then
      ! average flux
      do jy = 1, nyregrid
        datain = flxghg(:,jy,:)
        call average(nxregrid, n, timeghg, datain, ntstate, fluxes%time, dataout)
        fluxes%flx_nest(:,jy,:) = dataout(:,:)
      end do
    else if ( statres.lt.real(nt_flx/24) ) then
      ! interpolate flux
      do jy = 1, nyregrid
        datain = flxghg(:,jy,:)
        call interp(nxregrid, n, timeghg, datain, ntstate, fluxes%time, dataout)
        fluxes%flx_nest(:,jy,:) = dataout(:,:)
      end do
    else if ( statres.eq.real(nt_flx/24) ) then
      fluxes%flx_nest(:,:,:) = flxghg
    endif
    deallocate( datain )
    deallocate( dataout )
    deallocate( flxghg )
    deallocate( timeghg )

    ! ocean fluxes for lognormal case when inc_ocean is false
    if ( config%lognormal.and..not.config%inc_ocean ) then
      ! use land-sea mask to set all land fluxes to zero
      do n = 1, ntstate
        fluxes%flxocn(:,:,n) = abs(lsm(:,:) - 1.)*fluxes%flx_nest(:,:,n)
      end do
    endif

    ! nest -> use nested fluxes for state vector
    allocate( flx(nxregrid,nyregrid,ntstate) )
    flx(:,:,:) = fluxes%flx_nest

  else

    ! no nest -> use global fluxes for state vector 
    allocate( flx(nxgrid,nygrid,ntstate) )
    flx(:,:,:) = fluxes%flx

  endif

  ! prior state vector
  ! ------------------

  ! calculate area-weighted mean flux and uncertainty for each box

  if ( config%nested ) then
    ix0 = 1
    jy0 = 1
  else
    ix0 = minloc(gbl_lon,dim=1,mask=abs(gbl_lon-rllx).eq.minval(abs(gbl_lon-rllx)))
    jy0 = minloc(gbl_lat,dim=1,mask=abs(gbl_lat-rlly).eq.minval(abs(gbl_lat-rlly)))
  endif
  write(logid,*) 'init_ghg: ix0, jy0 = ',ix0,jy0

  flx_box(:,:) = 0.
  err_box(:,:) = 0.
  do n = 1, ntstate
    do ix = 1, nxregrid
      do jy = 1, nyregrid
        area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
        ! if inc_ocean flx_box contains all fluxes, if false only land fluxes
        if ( nbox_xy(ix,jy).gt.0 ) then
          flx_box(nbox_xy(ix,jy),n) = flx_box(nbox_xy(ix,jy),n) + &
                                        flx(ix0+ix-1,jy0+jy-1,n) * area
          err_box(nbox_xy(ix,jy),n) = err_box(nbox_xy(ix,jy),n) + &
                                        flx(ix0+ix-1,jy0+jy-1,n) * config%flxerr * area
        endif
      end do
    end do
    flx_box(:,n) = flx_box(:,n)/area_box
    err_box(:,n) = err_box(:,n)/area_box
  end do
  fluxes%flx_box(:,:,1) = flx_box(:,:)

  ! set minimum error
  do n = 1, ntstate
    do nb = 1, nbox
      err_box(nb,n) = max(config%flxerr_ll*numscale/3600.,err_box(nb,n))
    end do
  end do

  ! verbose output
  ! --------------

  if ( config%verbose ) then
    ! check prior fluxes read correctly (1-time dimension only)
    write(rowfmt,'(A,I6,A)') '(',nxregrid,'(E11.4,1X))'
    open(100,file=trim(files%path_output)//'flx_prior.txt',status='replace',action='write',iostat=ierr)
    do n = 1, nyregrid
      write(100,fmt=rowfmt) flx(ix0:(ix0+nxregrid-1),jy0+n-1,1)
    end do
    close(100)
    ! check regridding to variable resolution 
    write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
    open(100,file=trim(files%path_output)//'flx_box.txt',status='replace',action='write',iostat=ierr)
    do n = 1, ntstate
      write(100,fmt=rowfmt) flx_box(:,n)
    end do
    close(100)
  endif

  ! prior state vector
  ! ------------------

  ! number of state variables
  npvar = ntstate*nbox
  ndvar = nbox
  if ( config%opt_cini ) then
    nvar = npvar + ntcini*ncini
  else
    nvar = npvar
  endif
  write(logid,*) 'Number of flux state variables: ',npvar
  write(logid,*) 'Number of state variables per flux time step: ',ndvar
  write(logid,*) 'Total number of state variables: ',nvar

  ! allocate state and error vectors
  call alloc_states(states)

  if ( config%lognormal ) then
    ! lognormal distribution 
    ! optimize scalar z = ln(x/xb) 
    ! prior state vector z is zero
    states%px0(1:npvar) = 0.
  else
    ! normal distribution
    ! prior state vector zero since optimize offsets
     states%px0(1:npvar) = 0.
  endif

  ! assign prior initial mixing ratio scalars
  if ( config%opt_cini ) then
    if ( config%lognormal ) then
      states%px0(npvar+1:nvar) = 0.
    else
      states%px0(npvar+1:nvar) = 1.
    endif
  endif

  ! time stamp of flux variables
  states%xtime(:) = fluxes%time

  ! time stamp of initial mixing ratio scalars
  do n = 1, ntcini
    states%cinitime(n) = juldatei + dble((n-1)*cinires)
  end do
  print*, 'init_ghg: cinitime = ',states%cinitime

  if ( config%lognormal ) then
    ! lognormal 
    ! uncertainty of a scalar of fluxes -> use constant uncertainty
    ! transform to error of z
    xerr(:) = log(1.+config%flxerr)
  else
    ! normal
    ! average error over time steps
    xerr = abs(sum(err_box(:,:),dim=2)/real(ntstate) )
  endif

  ! assign best guess of state vector
  if ( (config%prior_bg.eq.1).and.((config%method.eq.'congrad').or.(config%method.eq.'m1qn3')) ) then
    ! from previous inversion
    call read_bg(files, config, states)
  else
    ! from prior
    states%px(:) = states%px0(:)
  endif

  ! testing
  open(100,file=trim(files%path_output)//'prior.txt',status='replace',action='write',iostat=ierr)
  do n = 1, nvar
    write(100,fmt='(E14.8)') states%px(n)
  end do
  close(100)  

  ! correlation and error covariance matrices
  ! -----------------------------------------

  corr(:,:) = 0.
  if ( config%spa_corr ) then
    write(logid,*) 'Using spatial correlation in B'
    ! check if file already exists (in case of rerun)
    inquire(file=trim(files%path_output)//'correl.txt',exist=lexist) 
    if ( lexist ) then
      write(logid,*) 'Reading spatial correlation from correl.txt'
      open(100,file=trim(files%path_output)//'correl.txt',action='read',status='old',iostat=ierr)
      write(rowfmt,'(A,I6,A)') '(',nbox,'(E11.4,1X))'
      print*, rowfmt
      do nb = 1, nbox
        read(100,fmt=rowfmt) corr(nb,:)
      end do
      close(100)
    else
      call correl(files, config, corr)
    endif
  else
    do n = 1, nbox
      corr(n,n) = 1.
    end do
  endif
  call error_cov(config, files, states, covar, corr, xerr)

  ! observations
  ! ------------

  call read_obs(config, files)
  call assign_obs(config, files, obs)

  ! initial concentrations
  ! ----------------------

  ! initial concentrations from termination of trajectories
  ! only do for a fresh run
  if ( ( config%restart.eq.0 ).and.( config%spec.ne.'bca') ) then  
    if ( files%init_type.lt.5 ) then
      ! daily or higher frequency
      call init_cini(files, config, obs)
    else
      ! monthly frequency
      call init_cini_month(files, config, obs)
    endif
  endif

  ! add random perturbation for MC
  ! ------------------------------

  if ( config%run_mode.eq.2 ) then
    write(logid,*) 'Adding random perturbation to state vector and observations...'
    ! perturb state vector
    call perturb_state(config, states, covar)
    ! perturb observations 
    call perturb_obs(config, obs)
  endif

  ! save datatype obs for restart
  ! -----------------------------

  if ( config%restart.eq.1 ) then
    ! read datatype obs from file
    write(logid,*) 'WARNING init_ghg: reading cini from previous run'
    open(100,file=trim(files%path_output)//'obsdatatype.txt',status='old',action='read',iostat=ierr)
    if ( ierr.ne.0 ) then
      write(logid,*) 'ERROR init_ghg: cannot open file obsdatatype.txt'
      stop
    endif
    allocate( recs(nobs) )
    allocate( jdate(nobs) )
    write(rowfmt,'(A,I1,A,I6,A)') '(A',recname_len,',1X,F14.6,1X,F10.4,1X,',ncini,'(F10.4,1X))'
    read(100,*) dump
    do n = 1, nobs
      read(100,rowfmt,iostat=ierr) recs(n), jdate(n), conc, obs%cini(n,:)
      if ( ierr.ne.0 ) then
        write(logid,*) 'ERROR init_ghg: reading obsdatatype.txt'
        exit
      endif
    end do
    ! check consistency
    if ( recs(1).ne.obs%recs(1).or.recs(nobs).ne.obs%recs(nobs).or.jdate(1).ne.obs%jdate(1).or.jdate(nobs).ne.obs%jdate(nobs) ) then
      write(logid,*) 'ERROR init_ghg: data in obsdatatype.txt inconsistent with run'
      print*, recs(1), obs%recs(1)
      print*, recs(nobs), obs%recs(nobs)
      print*, jdate(1), obs%jdate(1)
      print*, jdate(nobs), obs%jdate(nobs)
      stop
    endif
    deallocate( recs )
    deallocate( jdate )
  else
    ! save datatype obs
    call save_for_restart(files, obs)
  endif

end subroutine init_ghg


