!---------------------------------------------------------------------------------------
! FLEXINVERT: mod_save
!---------------------------------------------------------------------------------------
!  FLEXINVERT is free software: you can redistribute it and/or modify
!  it under the terms of the GNU General Public License as published by
!  the Free Software Foundation, either version 3 of the License, or
!  (at your option) any later version.
!
!  FLEXINVERT is distributed in the hope that it will be useful,
!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!  GNU General Public License for more details.
!
!  You should have received a copy of the GNU General Public License
!  along with FLEXINVERT.  If not, see <http://www.gnu.org/licenses/>.
!
!  Copyright 2017, Rona Thompson
!---------------------------------------------------------------------------------------
!
!> mod_save
!!
!! Purpose:    Module for saving the observation and state vector output.
!!
!---------------------------------------------------------------------------------------


module mod_save

  use mod_settings
  use mod_states
  use mod_var
  use mod_tools
  use netcdf

  implicit none
  private

  public :: save_state, save_co2, save_ghg, save_obs, save_nee_hourly, save_scalar_cini, &
              save_for_restart, save_average_fp, check

  contains

  ! --------------------------------------------------
  ! save_obs
  ! --------------------------------------------------
  !> save_obs
  !!
  !! Purpose:  Saves the observations to the file
  !!           monitor.txt
  !!
  ! --------------------------------------------------

  subroutine save_obs(files, config, obs)

    use mod_settings
    use mod_obs
    use mod_var
    use mod_dates

    type (files_t),    intent(in)   :: files
    type (config_t),   intent(in)   :: config
    type (obs_t),      intent(in)   :: obs

    character(len=max_path_len)     :: filename
    character(len=max_path_len)     :: rowfmt
    integer                         :: jjjjmmdd, hhmiss
    integer                         :: i, ierr
    real                            :: cinipri

    filename = trim(files%path_output)//'monitor.txt'
    open(100,file=trim(filename),status='replace',action='write',iostat=ierr)
    if ( ierr.ne.0 ) then 
      print*, 'ERROR: cannot write to '//trim(filename)
      stop
    endif

    if ( config%spec.eq.'co2' ) then
      ! CO2 species
      write(100,*) 'rec yyyymmdd hhmmss juldate conc cini cinipos bkg nee fff ocn prior post cpri cakpri diff error'
      do i = 1, nobs
        cinipri = sum(obs%cini(i,:))
        call caldate(obs%jdate(i), jjjjmmdd, hhmiss)
        write(rowfmt,'(A,I1,A,I6,A)') '(A',recname_len,',1X,I8,1X,I6,1X,F14.6,1X,',13,'(F10.4,1X))'
        print*, 'save_obs: ',rowfmt
        write(100,fmt=rowfmt) obs%recs(i), jjjjmmdd, hhmiss, obs%jdate(i), &
                obs%conc(i), cinipri, obs%cinipos(i), obs%bkg(i), &
                obs%nee(i), obs%fff(i), obs%ocn(i), obs%prior(i), &
                obs%model(i), obs%cpri(i), obs%cakpri(i), obs%delta(i), obs%err(i)
      end do
    else
      ! GHG species
      write(100,*) 'rec yyyymmdd hhmmss juldate conc cini cinipos bkg ghg prior post cpri cakpri diff error'
      write(rowfmt,'(A,I1,A,I6,A)') '(A',recname_len,',1X,I8,1X,I6,1X,F14.6,1X,',11,'(F10.4,1X))'
      do i = 1, nobs
        cinipri = sum(obs%cini(i,:))
        call caldate(obs%jdate(i), jjjjmmdd, hhmiss)
        ! if log-normal and ocean grid cells not optimized include ocean contribution in prior and posterior model
        ! in other case obs%ocn is zero
        write(100,fmt=rowfmt) obs%recs(i), jjjjmmdd, hhmiss, obs%jdate(i), &
                obs%conc(i), cinipri, obs%cinipos(i), obs%bkg(i), &
                obs%ghg(i), obs%prior(i)+obs%ocn(i), obs%model(i)+obs%ocn(i), &
                obs%cpri(i), obs%cakpri(i), obs%delta(i), obs%err(i)
      end do
    endif

    close(100)

  end subroutine save_obs

  ! --------------------------------------------------
  ! save_state
  ! --------------------------------------------------
  !> save_state
  !!
  !! Purpose:  Interface to save routines for state
  !!           vectors.
  !!
  ! --------------------------------------------------

  subroutine save_state(files, config, fluxes, states)

    use mod_settings
    use mod_fluxes
    use mod_states
    use mod_var

    type (files_t),    intent(in)   :: files
    type (config_t),   intent(in)   :: config
    type (fluxes_t),   intent(in)   :: fluxes
    type (states_t),   intent(in)   :: states
     
    if ( config%spec.eq.'co2' ) then
      call save_co2(files, config, fluxes, states)
      call save_nee_hourly(files, fluxes, states)
    else
      call save_ghg(files, config, fluxes, states)
    endif
    if ( config%opt_cini ) then
      call save_scalar_cini(files, config, states)
    endif

  end subroutine save_state

  ! --------------------------------------------------
  ! save_co2
  ! --------------------------------------------------
  !!> save_co2
  !!
  !! Purpose:  Saves the CO2 fluxes to the file
  !!           analysis.nc
  !!
  ! --------------------------------------------------

  subroutine save_co2(files, config, fluxes, states)

    use mod_settings
    use mod_fluxes
    use mod_states
    use mod_var
    use mod_dates
    use mod_strings

    type (files_t),    intent(in)              :: files
    type (config_t),   intent(in)              :: config
    type (fluxes_t),   intent(in)              :: fluxes
    type (states_t),   intent(in)              :: states

    character(len=*), parameter                :: fileout = 'analysis.nc'
    character(len=*), parameter                :: lonname = 'longitude'
    character(len=*), parameter                :: latname = 'latitude' 
    character(len=*), parameter                :: timename = 'time'
    character(len=*), parameter                :: ndtname = 'hours'
    character(len=*), parameter                :: pri_name = 'fpri'
    character(len=*), parameter                :: pos_name = 'fpos'
    character(len=*), parameter                :: xpos_name = 'fincrement'
    character(len=*), parameter                :: epri_name = 'epri'
    character(len=*), parameter                :: epos_name = 'epos'
    character(len=*), parameter                :: fff_name = 'fff'
    character(len=*), parameter                :: ocn_name = 'focn'
    character(len=*), parameter                :: units = 'units'
    character(len=*), parameter                :: dimunit = 'degrees'
    character(len=*), parameter                :: timeunit = 'julian_day'
    character(len=*), parameter                :: flxunit = 'kg/m2/s'
    character(len=*), parameter                :: ndtunit = 'hours'
    integer                                    :: ncid, lon_dimid, lat_dimid, time_dimid, ndt_dimid 
    integer, dimension(3)                      :: dimids
    integer, dimension(4)                      :: xpos_dimids
    integer                                    :: lon_varid, lat_varid, time_varid, ndt_varid
    integer                                    :: pri_varid, pos_varid, epri_varid, epos_varid, xpos_varid 
    integer                                    :: fff_varid, ocn_varid  
    integer                                    :: ix, jy, n, nt
    integer                                    :: flag, ni, ind
    integer                                    :: ix1, ix2, jy1, jy2
    real(kind=8), dimension(ntstate)           :: timeout
    real, dimension(ndt)                       :: ndtout
    real, dimension(nxregrid,nyregrid,ntstate) :: fpri, fpos, epri, epos
    real, dimension(ntstate)                   :: sumpri, sumpos
    real                                       :: area, landarea, totpri, totpos, totpri_post, totpos_post
    real, dimension(nxregrid,nyregrid,ntstate,ndt) :: xpos
    real, dimension(nxregrid,nyregrid,ntstate)     :: fff_reg, focn_reg
    real, dimension(:,:), allocatable              :: datain, dataout

    ! regrid fluxes and errors
    ! ------------------------

    fpri(:,:,:) = 0.
    fpos(:,:,:) = 0.
    xpos(:,:,:,:) = 0.
    epri(:,:,:) = 0.
    epos(:,:,:) = 0.
    do n = 1, ntstate
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).gt.0 ) then
            ! undo numerical scaling and regrid
            ! average fluxes and errors over sub-daily time steps
            do nt = 1, ndt
              fpri(ix,jy,n) = fpri(ix,jy,n) + states%px0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))/real(ndt)/numscale
              fpos(ix,jy,n) = fpos(ix,jy,n) + states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))/real(ndt)/numscale
              epri(ix,jy,n) = epri(ix,jy,n) + states%pxerr0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))/real(ndt)/numscale
              epos(ix,jy,n) = epos(ix,jy,n) + states%pxerr((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))/real(ndt)/numscale
              xpos(ix,jy,n,nt) = states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))/numscale
            end do
          endif
        end do
      end do
    end do

    ! redistribute so that flux increments on ocean boxes are attributed to 
    ! land grid cells within the same aggregated box
    ! note: future improvement to define aggregated grid with clear separation along coastlines
    ! then this step no longer needed and should be more accurate!
    totpri = 0.
    totpos = 0.
    totpri_post = 0.
    totpos_post = 0.
    do n = 1, nbox
      sumpri(:) = 0.
      sumpos(:) = 0.
      landarea = 0.
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).eq.n ) then
            area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx) 
            if ( lsm(ix,jy).ge.0.5 ) then
              landarea = landarea + area
            endif
            sumpri(:) = sumpri(:) + fpri(ix,jy,:)*area
            sumpos(:) = sumpos(:) + fpos(ix,jy,:)*area
          endif
        end do
      end do
      ! testing
      if(.not.isnan(sum(sumpri))) totpri = totpri + sum(sumpri)/real(ntstate)
      if(.not.isnan(sum(sumpos))) totpos = totpos + sum(sumpos)/real(ntstate)
      ! end testing
      if (landarea > 0) then
        sumpri(:) = sumpri(:)/landarea
        sumpos(:) = sumpos(:)/landarea
      else
        sumpri(:) = 0.
        sumpos(:) = 0.
      endif
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).eq.n .and. lsm(ix,jy).ge.0.5 ) then
            ! land so assign redistributed NEE increment
            fpri(ix,jy,:) = sumpri(:)
            fpos(ix,jy,:) = sumpos(:)
          else if ( nbox_xy(ix,jy).eq.n ) then
            ! ocean so NEE increment zero
            fpri(ix,jy,:) = 0.
            fpos(ix,jy,:) = 0.
          endif
        end do
      end do
    end do
    ! testing    
    do jy = 1, nyregrid
      area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
      do ix = 1, nxregrid
        totpri_post = totpri_post + sum(fpri(ix,jy,:))*area/real(ntstate)
        totpos_post = totpos_post + sum(fpos(ix,jy,:))*area/real(ntstate)
      enddo
    end do
    ! testing
    print*, 'Prior total before redistribution = ',totpri*3600.*24.*365./1.e9
    print*, 'Prior total after redistribution = ',totpri_post*3600.*24.*365./1.e9
    print*, 'Post total before redistribution = ',totpos*3600.*24.*365./1.e9
    print*, 'Post total after redistribution = ',totpos_post*3600.*24.*365./1.e9

    ! add prior NEE
    ! -------------

    allocate( datain(nxregrid,nday*ndgrid) )
    allocate( dataout(nxregrid,ntstate) )
    do jy = 1, nyregrid
      ! average to state vector timestamp
      datain = fluxes%flxnee_nest(:,jy,:)
      if ((nday*ndgrid).gt.1.and.ntstate.gt.1) then
        call average(nxregrid, nday*ndgrid, fluxes%timefp, datain, ntstate, states%xtime, dataout)
      else if ((nday*ndgrid).gt.1) then
        n = minloc(fluxes%timefp,dim=1,mask=abs(fluxes%timefp-states%xtime(1)).eq.minval(abs(fluxes%timefp-states%xtime(1))))
        dataout(:,ntstate) = datain(:,n)
      endif
      fpri(:,jy,:) = fpri(:,jy,:) + dataout(:,:)/numscale
      fpos(:,jy,:) = fpos(:,jy,:) + dataout(:,:)/numscale
    end do
    print*, 'completed NEE'
    deallocate( datain )
    deallocate( dataout )

    ! ocean fluxes
    ! ------------

    ! low temporal resolution fluxes
    ix1 = int((rllx - llx + dx)/dx)
    ix2 = int((rurx - llx)/dx)
    jy1 = int((rlly - lly + dy)/dy)
    jy2 = int((rury - lly)/dy)
    if ( config%nested ) then
      focn_reg(:,:,:) = fluxes%flxocn_nest(:,:,:)/numscale
    else
      focn_reg(:,:,:) = fluxes%flxocn(ix1:ix2,jy1:jy2,:)/numscale
    endif

    ! fossil fuel
    ! -----------

    allocate( datain(nxregrid,nday*ndgrid) )
    allocate( dataout(nxregrid,ntstate) )
    do jy = 1, nyregrid
      ! average to state vector timestamp
      datain = fluxes%flxff_nest(:,jy,:)
      if ((nday*ndgrid).gt.1.and.ntstate.gt.1) then
        call average(nxregrid, nday*ndgrid, fluxes%timefp, datain, ntstate, states%xtime, dataout)
      else if ((nday*ndgrid).gt.1) then
        n = minloc(fluxes%timefp,dim=1,mask=abs(fluxes%timefp-states%xtime(1)).eq.minval(abs(fluxes%timefp-states%xtime(1))))
        dataout(:,ntstate) = datain(:,n)
      endif
      fff_reg(:,jy,:) = dataout(:,:)/numscale
    end do
    print*, 'completed fossil fuel'
    deallocate( datain )
    deallocate( dataout )
    
    ! write ncdf file
    ! ---------------
 
    ! adjust time to days since 1-Jan-1900
    timeout = states%xtime - juldate(19000101, 0)

    ! sub-daily time dimension
    do nt = 1, ndt
      ndtout(nt) = (nt-1)*24./ndt
    end do

    call check( nf90_create(trim(files%path_output)//fileout, nf90_clobber, ncid) )

    call check( nf90_def_dim(ncid, lonname, nxregrid, lon_dimid) )
    call check( nf90_def_var(ncid, lonname, nf90_real, lon_dimid, lon_varid) )
    call check( nf90_put_att(ncid, lon_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, latname, nyregrid, lat_dimid) )
    call check( nf90_def_var(ncid, latname, nf90_real, lat_dimid, lat_varid) )
    call check( nf90_put_att(ncid, lat_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, timename, ntstate, time_dimid) )
    call check( nf90_def_var(ncid, timename, nf90_real, time_dimid, time_varid) )
    call check( nf90_put_att(ncid, time_varid, units, timeunit) )
    call check( nf90_def_dim(ncid, ndtname, ndt, ndt_dimid) )
    call check( nf90_def_var(ncid, ndtname, nf90_real, ndt_dimid, ndt_varid) )
    call check( nf90_put_att(ncid, ndt_varid, units, ndtunit) )

    dimids = (/ lon_dimid, lat_dimid, time_dimid /)
    xpos_dimids = (/ lon_dimid, lat_dimid, time_dimid, ndt_dimid /)

    call check( nf90_def_var(ncid, pri_name, nf90_real, dimids, pri_varid) )
    call check( nf90_put_att(ncid, pri_varid, units, flxunit) )
    call check( nf90_def_var(ncid, epri_name, nf90_real, dimids, epri_varid) )
    call check( nf90_put_att(ncid, epri_varid, units, flxunit) )
    call check( nf90_def_var(ncid, pos_name, nf90_real, dimids, pos_varid) )
    call check( nf90_put_att(ncid, pos_varid, units, flxunit) )
    call check( nf90_def_var(ncid, epos_name, nf90_real, dimids, epos_varid) )
    call check( nf90_put_att(ncid, epos_varid, units, flxunit) )
    call check( nf90_def_var(ncid, ocn_name, nf90_real, dimids, ocn_varid) )
    call check( nf90_put_att(ncid, ocn_varid, units, flxunit) )
    call check( nf90_def_var(ncid, fff_name, nf90_real, dimids, fff_varid) )
    call check( nf90_put_att(ncid, fff_varid, units, flxunit) )
    call check( nf90_def_var(ncid, xpos_name, nf90_real, xpos_dimids, xpos_varid) )
    call check( nf90_put_att(ncid, xpos_varid, units, flxunit) )

    call check( nf90_enddef(ncid) )

    call check( nf90_put_var(ncid, lon_varid, reg_lon) )
    call check( nf90_put_var(ncid, lat_varid, reg_lat) )
    call check( nf90_put_var(ncid, time_varid, timeout) )
    call check( nf90_put_var(ncid, ndt_varid, ndtout) )

    call check( nf90_put_var(ncid, pri_varid, fpri) )
    call check( nf90_put_var(ncid, epri_varid, epri) )
    call check( nf90_put_var(ncid, pos_varid, fpos) )
    call check( nf90_put_var(ncid, epos_varid, epos) )
    call check( nf90_put_var(ncid, ocn_varid, focn_reg) )
    call check( nf90_put_var(ncid, fff_varid, fff_reg) )
    call check( nf90_put_var(ncid, xpos_varid, xpos) )

    call check( nf90_close(ncid) )

  end subroutine save_co2

  ! --------------------------------------------------
  ! save_ghg
  ! --------------------------------------------------
  !> save_ghg
  !!
  !! Purpose:  Saves the GHG fluxes to the file
  !!           analysis.nc
  !!
  ! --------------------------------------------------

  subroutine save_ghg(files, config, fluxes, states)

    use mod_settings
    use mod_fluxes
    use mod_states
    use mod_dates
    use mod_var

    type (files_t),    intent(in)              :: files
    type (config_t),   intent(in)              :: config
    type (fluxes_t),   intent(in)              :: fluxes
    type (states_t),   intent(in)              :: states

    character(len=*), parameter                :: filename = 'analysis.nc'
    character(len=*), parameter                :: lonname = 'longitude'
    character(len=*), parameter                :: latname = 'latitude'
    character(len=*), parameter                :: timename = 'time'
    character(len=*), parameter                :: pri_name = 'fpri'
    character(len=*), parameter                :: pos_name = 'fpos'
    character(len=*), parameter                :: xpos_name = 'fincrement'
    character(len=*), parameter                :: epri_name = 'epri'
    character(len=*), parameter                :: epos_name = 'epos'
    character(len=*), parameter                :: units = 'units'
    character(len=*), parameter                :: dimunit = 'degrees'
    character(len=*), parameter                :: timeunit = 'julian_day'
    character(len=*), parameter                :: flxunit = 'kg/m2/s'
    real,             parameter                :: smallnum = 1.e-15
    integer                                    :: ncid, lon_dimid, lat_dimid, time_dimid
    integer, dimension(3)                      :: dimids
    integer                                    :: lon_varid, lat_varid, time_varid
    integer                                    :: pri_varid, pos_varid, epri_varid, epos_varid, xpos_varid
    integer                                    :: ix, jy, n, ix0, jy0, nb
    real(kind=8), dimension(ntstate)           :: timeout
    real, dimension(nxregrid,nyregrid,ntstate) :: fpri, fpos, epri, epos, xpos, focn_reg
    real                                       :: area, totpos

    ! regrid fluxes and errors
    ! ------------------------

    ix0 = minloc(gbl_lon,dim=1,mask=abs(gbl_lon-rllx).eq.minval(abs(gbl_lon-rllx)))
    jy0 = minloc(gbl_lat,dim=1,mask=abs(gbl_lat-rlly).eq.minval(abs(gbl_lat-rlly)))
    write(logid,*) 'save_ghg: ix0, jy0 = ',ix0, jy0

    fpri(:,:,:) = 0.
    fpos(:,:,:) = 0.
    xpos(:,:,:) = 0.
    epri(:,:,:) = 0.
    epos(:,:,:) = 0.
    do n = 1, ntstate
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( config%nested ) then
            fpri(ix,jy,n) = fluxes%flx_nest(ix,jy,n)/numscale
          else
            fpri(ix,jy,n) = fluxes%flx(ix0+ix-1,jy0+jy-1,n)/numscale
          endif
          ! if inc_ocean is true state vector includes all fluxes, if false need to add ocean fluxes
          if ( nbox_xy(ix,jy).gt.0 ) then
            if ( config%nested ) then
              ! nested domain
              if ( config%lognormal ) then
                ! lognormal: optimize scalars of fluxes
                fpos(ix,jy,n) = exp(states%px((n-1)*nbox+nbox_xy(ix,jy))) * fluxes%flx_nest(ix,jy,n)/numscale
              else
                ! normal distribution: optimize offsets
                if ( fluxes%flx_box(nbox_xy(ix,jy),n,1).gt.smallnum ) then
                  ! regrid flux increments according to prior flux distribution and add prior flux
                  fpos(ix,jy,n) = states%px((n-1)*nbox+nbox_xy(ix,jy)) *  &
                                  fluxes%flx_nest(ix,jy,n)/fluxes%flx_box(nbox_xy(ix,jy),n,1)/numscale + &
                                  fpri(ix,jy,n)
                else
                  ! prior in box very small so no redistribution
                  fpos(ix,jy,n) = states%px((n-1)*nbox+nbox_xy(ix,jy))/numscale + fpri(ix,jy,n)
                endif
              endif ! normal/lognormal
            else
              ! no nested domain
              if ( config%lognormal ) then
                ! lognormal: optimize scalars of fluxes
                fpos(ix,jy,n) = exp(states%px((n-1)*nbox+nbox_xy(ix,jy))) * fluxes%flx(ix0+ix-1,jy0+jy-1,n)/numscale
              else
                ! normal distribution: optimize offsets
                if ( fluxes%flx_box(nbox_xy(ix,jy),n,1).gt.smallnum ) then
                  ! regrid flux increments according to prior flux distribution and add prior flux
                  fpos(ix,jy,n) = states%px((n-1)*nbox+nbox_xy(ix,jy)) *  &
                                  fluxes%flx(ix0+ix-1,jy0+jy-1,n)/fluxes%flx_box(nbox_xy(ix,jy),n,1)/numscale + &
                                  fpri(ix,jy,n)
                else
                  ! prior in box very small so no redistribution
                  fpos(ix,jy,n) = states%px((n-1)*nbox+nbox_xy(ix,jy))/numscale + fpri(ix,jy,n)
                endif
              endif ! normal/lognormal
            endif ! nested
            ! regrid without adding prior fluxes
            if ( config%lognormal ) then
              ! lognormal distribution
              ! zerr = ln(1 + xerr) so xerr = exp(zerr) - 1 which is fraction of the flux
              if ( config%nested ) then
                epri(ix,jy,n) = fluxes%flx_nest(ix,jy,n)*abs(exp(states%pxerr0((n-1)*nbox+nbox_xy(ix,jy))) - 1.)/numscale
                epos(ix,jy,n) = fluxes%flx_nest(ix,jy,n)*abs(exp(states%pxerr((n-1)*nbox+nbox_xy(ix,jy))) - 1.)/numscale
              else
                epri(ix,jy,n) = fluxes%flx(ix0+ix-1,jy0+jy-1,n)*abs(exp(states%pxerr0((n-1)*nbox+nbox_xy(ix,jy))) - 1.)/numscale
                epos(ix,jy,n) = fluxes%flx(ix0+ix-1,jy0+jy-1,n)*abs(exp(states%pxerr((n-1)*nbox+nbox_xy(ix,jy))) - 1.)/numscale
              endif
              xpos(ix,jy,n) = exp(states%px((n-1)*nbox+nbox_xy(ix,jy)))
            else
              ! normal distribution
              epri(ix,jy,n) = states%pxerr0((n-1)*nbox+nbox_xy(ix,jy))/numscale
              epos(ix,jy,n) = states%pxerr((n-1)*nbox+nbox_xy(ix,jy))/numscale
              xpos(ix,jy,n) = states%px((n-1)*nbox+nbox_xy(ix,jy))/numscale
            endif
          else
            ! include ocean fluxes if not optimized
            if ( config%nested ) then
              fpos(ix,jy,n) = fluxes%flx_nest(ix,jy,n)/numscale
            else 
              fpos(ix,jy,n) = fluxes%flx(ix0+ix-1,jy0+jy-1,n)/numscale
            endif
          endif ! nbox_xy
        end do
      end do
    end do

    ! testing
    totpos = 0.
    do n = 1, ntstate
      do nb = 1, nbox
        if ( config%lognormal ) then
          totpos = totpos + exp(states%px((n-1)*nbox+nb))*fluxes%flx_box(nb,n,1)*area_box(nb)/numscale
        else
          totpos = totpos + states%px((n-1)*nbox+nb)*area_box(nb)/numscale
        endif
      end do
    end do
    totpos = totpos/real(ntstate)
    print*, 'Total increment before regridding = ',totpos*3600.*24.*365./1.e9
    totpos = 0.
    do jy = 1, nyregrid
      area = grid_area(reg_lat(jy)+0.5*rdy, rdy, rdx)
      do ix = 1, nxregrid
        if ( config%lognormal ) then
          totpos = totpos + sum(fpos(ix,jy,:))*area/real(ntstate)
        else
          totpos = totpos + sum(fpos(ix,jy,:)-fpri(ix,jy,:))*area/real(ntstate)
        endif
      end do
    end do
    print*, 'Total increment after regridding = ',totpos*3600.*24.*365./1.e9

    ! write ncdf file
    ! ---------------

    ! adjust time to days since 1-Jan-1900
    timeout = states%xtime - juldate(19000101, 0)

    call check( nf90_create(trim(files%path_output)//filename, nf90_clobber, ncid) )

    call check( nf90_def_dim(ncid, lonname, nxregrid, lon_dimid) )
    call check( nf90_def_var(ncid, lonname, nf90_real, lon_dimid, lon_varid) )
    call check( nf90_put_att(ncid, lon_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, latname, nyregrid, lat_dimid) )
    call check( nf90_def_var(ncid, latname, nf90_real, lat_dimid, lat_varid) )
    call check( nf90_put_att(ncid, lat_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, timename, ntstate, time_dimid) )
    call check( nf90_def_var(ncid, timename, nf90_real, time_dimid, time_varid) )
    call check( nf90_put_att(ncid, time_varid, units, timeunit) )

    dimids = (/ lon_dimid, lat_dimid, time_dimid /)

    call check( nf90_def_var(ncid, pri_name, nf90_real, dimids, pri_varid) )
    call check( nf90_put_att(ncid, pri_varid, units, flxunit) )
    call check( nf90_def_var(ncid, epri_name, nf90_real, dimids, epri_varid) )
    call check( nf90_put_att(ncid, epri_varid, units, flxunit) )
    call check( nf90_def_var(ncid, pos_name, nf90_real, dimids, pos_varid) )
    call check( nf90_put_att(ncid, pos_varid, units, flxunit) )
    call check( nf90_def_var(ncid, epos_name, nf90_real, dimids, epos_varid) )
    call check( nf90_put_att(ncid, epos_varid, units, flxunit) )
    call check( nf90_def_var(ncid, xpos_name, nf90_real, dimids, xpos_varid) )
    call check( nf90_put_att(ncid, xpos_varid, units, flxunit) )

    call check( nf90_enddef(ncid) )

    call check( nf90_put_var(ncid, lon_varid, reg_lon) )
    call check( nf90_put_var(ncid, lat_varid, reg_lat) )
    call check( nf90_put_var(ncid, time_varid, timeout) )

    call check( nf90_put_var(ncid, pri_varid, fpri) )
    call check( nf90_put_var(ncid, epri_varid, epri) )
    call check( nf90_put_var(ncid, pos_varid, fpos) )
    call check( nf90_put_var(ncid, epos_varid, epos) )
    call check( nf90_put_var(ncid, xpos_varid, xpos) )

    call check( nf90_close(ncid) )

  end subroutine save_ghg

  ! --------------------------------------------------
  ! save_nee_hourly
  ! --------------------------------------------------
  !> save_nee_hourly
  !!
  !! Purpose:  Saves the NEE fluxes at sub-daily
  !!           time steps (same resolution as prior
  !!           NEE) to the file analysis_nee.nc.
  !!
  ! --------------------------------------------------

  subroutine save_nee_hourly(files, fluxes, states)

    use mod_settings
    use mod_fluxes
    use mod_states
    use mod_var
    use mod_dates
    use mod_strings

    type (files_t),    intent(in)                     :: files
    type (fluxes_t),   intent(in)                     :: fluxes
    type (states_t),   intent(in)                     :: states

    character(len=*), parameter                       :: fileout = 'analysis_nee.nc'
    character(len=*), parameter                       :: lonname = 'longitude'
    character(len=*), parameter                       :: latname = 'latitude'
    character(len=*), parameter                       :: timename = 'days'
    character(len=*), parameter                       :: hourname = 'hours'
    character(len=*), parameter                       :: pri_name = 'nee_pri'
    character(len=*), parameter                       :: pos_name = 'nee_pos'
    character(len=*), parameter                       :: xpos_name = 'nee_increment'
    character(len=*), parameter                       :: units = 'units'
    character(len=*), parameter                       :: dimunit = 'degrees'
    character(len=*), parameter                       :: timeunit = 'julian_day'
    character(len=*), parameter                       :: hourunit = 'hour'
    character(len=*), parameter                       :: flxunit = 'kg/m2/s'
    integer                                           :: ncid, lon_dimid, lat_dimid, time_dimid, hour_dimid
    integer, dimension(4)                             :: dimids
    integer                                           :: lon_varid, lat_varid, time_varid, hour_varid
    integer                                           :: pri_varid, pos_varid, xpos_varid
    integer                                           :: ix, jy, n, nd, nn, nt, d 
    real(kind=8)                                      :: jd
    real                                              :: hh, hutc
    integer, dimension(ndgrid)                        :: hours
    real(kind=8), dimension(nday)                     :: time, timeout
    real, dimension(nxregrid,nyregrid,nday,ndgrid)    :: fpri, fpos, xpos
    real, dimension(nxregrid,nyregrid,ndgrid)         :: flx_nee

    do nd = 1, ndgrid
      hours(nd) = (nd-1)*24/ndgrid
    end do

    ! interpolate flux increments and add diurnal cycle 
    ! -------------------------------------------------

    fpri(:,:,:,:) = 0.
    fpos(:,:,:,:) = 0.
    xpos(:,:,:,:) = 0.
    time(:) = 0d0

    jd = juldatei
    d = 0
    do while (jd.le.juldatef)
      d = d + 1
      n = int(floor((jd - juldatei)/real(statres,kind=8))) + 1
      n = min(n,ntstate)
      ! flux NEE current day
      flx_nee = fluxes%flxnee_nest(:,:,(d-1)*ndgrid+1:d*ndgrid)
      do jy = 1, nyregrid
        do ix = 1, nxregrid
          if ( nbox_xy(ix,jy).le.0 ) cycle
          ! loop over number of flexpart timesteps per day
          do nd = 1, ndgrid
            hutc = real(nd - 1)*24./real(ndgrid)
            hh = hutc + hloc(nbox_xy(ix,jy))
            if ( hh.lt.0 ) then
              hh = hh + 24
             else if ( hh.ge.24 ) then
              hh = hh - 24
            endif
            ! note: ndt is number of sub-daily timesteps in state vector
            ! find index nt to sub-daily timestep in state vector
            do nn = 1, ndt 
              if ( (hh.ge.(nn-1)*24/ndt).and.(hh.lt.nn*24/ndt) ) nt = nn
            end do
            if ( hh.gt.((nt-1)*24/ndt+24/ndt/2).and.nt.lt.ndt ) then
              ! interpolate between current and next timesteps
              fpri(ix,jy,d,nd) = states%px0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px0((n-1)*ndt*nbox+nt*nbox+nbox_xy(ix,jy)) - &
                                 states%px0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-1)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                 flx_nee(ix,jy,nd)
              fpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px((n-1)*ndt*nbox+nt*nbox+nbox_xy(ix,jy)) - &
                                 states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-1)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                 flx_nee(ix,jy,nd)
              xpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px((n-1)*ndt*nbox+nt*nbox+nbox_xy(ix,jy)) - &
                                 states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-1)*24/ndt - 24/ndt/2)*real(ndt)/24.
            else if ( hh.lt.((nt-1)*24/ndt+24/ndt/2).and.nt.gt.1) then
              ! interpolate between previous and current timesteps 
              fpri(ix,jy,d,nd) = states%px0((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) - &
                                 states%px0((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-2)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                 flx_nee(ix,jy,nd)
              fpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) - &
                                 states%px((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-2)*24/ndt - 24/ndt/2)*real(ndt)/24. + &
                                 flx_nee(ix,jy,nd)
              xpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy)) + &
                                 (states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) - &
                                 states%px((n-1)*ndt*nbox+(nt-2)*nbox+nbox_xy(ix,jy))) * &
                                 (hh - (nt-2)*24/ndt - 24/ndt/2)*real(ndt)/24.
            else
              ! take current timestep
              fpri(ix,jy,d,nd) = states%px0((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                 flx_nee(ix,jy,nd)
              fpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy)) + &
                                 flx_nee(ix,jy,nd)
              xpos(ix,jy,d,nd) = states%px((n-1)*ndt*nbox+(nt-1)*nbox+nbox_xy(ix,jy))
            endif
          end do
        end do
      end do
      time(d) = jd
      jd = jd + 1d0
    end do
    fpri(:,:,:,:) = fpri(:,:,:,:)/numscale
    fpos(:,:,:,:) = fpos(:,:,:,:)/numscale
    xpos(:,:,:,:) = xpos(:,:,:,:)/numscale

    ! write hourly NEE
    ! ----------------

    ! adjust time to days since 1-Jan-1900
    timeout = time - juldate(19000101, 0) 

    call check( nf90_create(trim(files%path_output)//fileout, nf90_clobber, ncid) )

    call check( nf90_def_dim(ncid, lonname, nxregrid, lon_dimid) )
    call check( nf90_def_var(ncid, lonname, nf90_real, lon_dimid, lon_varid) )
    call check( nf90_put_att(ncid, lon_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, latname, nyregrid, lat_dimid) )
    call check( nf90_def_var(ncid, latname, nf90_real, lat_dimid, lat_varid) )
    call check( nf90_put_att(ncid, lat_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, timename, nday, time_dimid) )
    call check( nf90_def_var(ncid, timename, nf90_real, time_dimid, time_varid) )
    call check( nf90_put_att(ncid, time_varid, units, timeunit) )
    call check( nf90_def_dim(ncid, hourname, ndgrid, hour_dimid) )
    call check( nf90_def_var(ncid, hourname, nf90_real, hour_dimid, hour_varid) )
    call check( nf90_put_att(ncid, hour_varid, units, hourunit) )

    dimids = (/ lon_dimid, lat_dimid, time_dimid, hour_dimid /)

    call check( nf90_def_var(ncid, pri_name, nf90_real, dimids, pri_varid) )
    call check( nf90_put_att(ncid, pri_varid, units, flxunit) )
    call check( nf90_def_var(ncid, pos_name, nf90_real, dimids, pos_varid) )
    call check( nf90_put_att(ncid, pos_varid, units, flxunit) )
    call check( nf90_def_var(ncid, xpos_name, nf90_real, dimids, xpos_varid) )
    call check( nf90_put_att(ncid, xpos_varid, units, flxunit) )

    call check( nf90_enddef(ncid) )

    call check( nf90_put_var(ncid, lon_varid, reg_lon) )
    call check( nf90_put_var(ncid, lat_varid, reg_lat) )
    call check( nf90_put_var(ncid, time_varid, real(timeout,kind=4)) )
    call check( nf90_put_var(ncid, hour_varid, hours) )
    call check( nf90_put_var(ncid, pri_varid, fpri) )
    call check( nf90_put_var(ncid, pos_varid, fpos) )
    call check( nf90_put_var(ncid, xpos_varid, xpos) )

    call check( nf90_close(ncid) )

  end subroutine save_nee_hourly

  ! --------------------------------------------------
  ! save_scalar_cini
  ! --------------------------------------------------
  !> save_scalar_cini
  !!
  !! Purpose:  Saves the scalars of the initial
  !!           mixing ratios and their uncertainties
  !!
  ! --------------------------------------------------

  subroutine save_scalar_cini(files, config, states)

    use mod_settings
    use mod_states
    use mod_var

    type (files_t),  intent(in)             :: files
    type (config_t), intent(in)             :: config
    type (states_t), intent(in)             :: states
    character(len=max_path_len)             :: rowfmt
    integer                                 :: n, ierr

    ! write to file
    open(100,file=trim(files%path_output)//'scalars_cini.txt',status='replace',action='write',iostat=ierr)
    write(rowfmt,'(A,I6,A)') '(',4,'(E11.4,1X))'
    write(100,*) 'prior, pri_err, post, pos_err'
    do n = 1, ntcini*ncini
      if (config%lognormal) then
        ! for lognormal these are the log transformed scalars
        ! so undo the log transformation here 
        write(100,rowfmt) exp(states%px0(npvar+n)), exp(states%pxerr0(npvar+n)), exp(states%px(npvar+n)), exp(states%pxerr(npvar+n))
      else
        write(100,rowfmt) states%px0(npvar+n), states%pxerr0(npvar+n), states%px(npvar+n), states%pxerr(npvar+n)
      endif
    end do
    close(100)

  end subroutine save_scalar_cini

  ! --------------------------------------------------
  ! save_for_restart
  ! --------------------------------------------------
  !> save_for_restart
  !!
  !! Purpose:  Saves the datatype obs for  
  !!           eventual restart
  !!
  ! --------------------------------------------------

  subroutine save_for_restart(files, obs)

    use mod_settings
    use mod_obs
    use mod_var

    type (files_t),  intent(in)             :: files
    type (obs_t),    intent(in)             :: obs
    character(len=max_path_len)             :: rowfmt
    integer                                 :: i, ierr

    ! write to file
    open(100,file=trim(files%path_output)//'obsdatatype.txt',status='replace',action='write',iostat=ierr)
    write(rowfmt,'(A,I1,A,I6,A)') '(A',recname_len,',1X,F14.6,1X,F10.4,1X,',ncini,'(F10.4,1X))'
    write(100,*) 'recs jdate conc cini'
    do i = 1, nobs
      write(100,rowfmt) obs%recs(i), obs%jdate(i), obs%conc(i), obs%cini(i,:)
    end do
    close(100)

  end subroutine save_for_restart

  ! --------------------------------------------------
  ! save_average_fp
  ! --------------------------------------------------
  !> save_average_fp
  !!
  !! Purpose:  Save averaged footprints
  !!
  ! --------------------------------------------------

  subroutine save_average_fp(files, filename, nx_in, ny_in, nt_in, ngrid, grid, lon_in, lat_in, gtime)

    use mod_settings
    use mod_fluxes
    use mod_states
    use mod_var
    use mod_dates
    use mod_strings

    type (files_t),                    intent(in)     :: files
    character(len=max_path_len),       intent(in)     :: filename
    integer,                           intent(in)     :: nx_in, ny_in, nt_in, ngrid
    real, dimension(nx_in,ny_in,nt_in),intent(in)     :: grid
    real, dimension(nx_in),            intent(in)     :: lon_in
    real, dimension(ny_in),            intent(in)     :: lat_in
    real(kind=8), dimension(nt_in),    intent(in)     :: gtime

    character(len=*), parameter                       :: lonname = 'longitude'
    character(len=*), parameter                       :: latname = 'latitude'
    character(len=*), parameter                       :: timename = 'time'
    character(len=*), parameter                       :: varname = 'srr'
    character(len=*), parameter                       :: units = 'units'
    character(len=*), parameter                       :: dimunit = 'degrees'
    character(len=*), parameter                       :: varunit = 's m3 kg^-1'
    integer                                           :: ncid, lon_dimid, lat_dimid, time_dimid
    integer, dimension(3)                             :: dimids
    integer                                           :: varid, lon_varid, lat_varid, time_varid

    call check( nf90_create(trim(files%path_flexncdf)//filename, nf90_clobber, ncid) )

    call check( nf90_def_dim(ncid, lonname, nx_in, lon_dimid) )
    call check( nf90_def_var(ncid, lonname, nf90_real, lon_dimid, lon_varid) )
    call check( nf90_put_att(ncid, lon_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, latname, ny_in, lat_dimid) )
    call check( nf90_def_var(ncid, latname, nf90_real, lat_dimid, lat_varid) )
    call check( nf90_put_att(ncid, lat_varid, units, dimunit) )
    call check( nf90_def_dim(ncid, timename, ngrid, time_dimid) )
    call check( nf90_def_var(ncid, timename, nf90_double, time_dimid, time_varid) )

    dimids = (/ lon_dimid, lat_dimid, time_dimid /)

    call check( nf90_def_var(ncid, varname, nf90_real, dimids, varid) )
    call check( nf90_put_att(ncid, varid, units, varunit) )

    call check( nf90_enddef(ncid) )

    call check( nf90_put_var(ncid, lon_varid, lon_in) )
    call check( nf90_put_var(ncid, lat_varid, lat_in) )
    print*, 'save_average_fp: gtime = ',gtime
    call check( nf90_put_var(ncid, time_varid, gtime(1:ngrid)) )
    call check( nf90_put_var(ncid, varid, grid(:,:,1:ngrid)) )

    call check( nf90_close(ncid) )

  end subroutine save_average_fp

  ! --------------------------------------------------
  ! check
  ! --------------------------------------------------

  subroutine check(status)

    integer, intent(in) :: status
    
    if ( status.ne.nf90_noerr ) then 
      write(logid,*) trim(nf90_strerror(status))
      stop
    end if

  end subroutine check  

  ! --------------------------------------------------


end module mod_save


